﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.EGO {
  [Item("VariableVariabilityAnalyzer", "Analyzes the correlation between perdictions and actual fitness values")]
  [StorableClass]
  public class VariableVariabilityAnalyzer : SingleSuccessorOperator, IAnalyzer {
    public override bool CanChangeName => true;
    public bool EnabledByDefault => false;

    public ILookupParameter<ModifiableDataset> DatasetParameter => (ILookupParameter<ModifiableDataset>)Parameters["Dataset"];
    public ILookupParameter<ResultCollection> ResultsParameter => (ILookupParameter<ResultCollection>)Parameters["Results"];
    public ILookupParameter<IntValue> InitialEvaluationsParameter => (ILookupParameter<IntValue>)Parameters["Initial Evaluations"];
    public IFixedValueParameter<IntValue> LookBackSizeParameter => (IFixedValueParameter<IntValue>)Parameters["LookBackSize"];

    private const string NormalizedPlotName = "Normalized Variable Variance";
    private const string PlotName = "Variable Variance";

    [StorableConstructor]
    protected VariableVariabilityAnalyzer(bool deserializing) : base(deserializing) { }
    protected VariableVariabilityAnalyzer(VariableVariabilityAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public VariableVariabilityAnalyzer() {
      Parameters.Add(new FixedValueParameter<IntValue>("LookBackSize", new IntValue(10)));
      Parameters.Add(new LookupParameter<IntValue>("Initial Evaluations"));
      Parameters.Add(new LookupParameter<ModifiableDataset>("Dataset"));
      Parameters.Add(new LookupParameter<ResultCollection>("Results", "The collection to store the results in."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new VariableVariabilityAnalyzer(this, cloner);
    }

    public sealed override IOperation Apply() {
      var dataset = DatasetParameter.ActualValue;
      var results = ResultsParameter.ActualValue;
      var initialEvals = InitialEvaluationsParameter.ActualValue.Value;
      var lbsize = LookBackSizeParameter.Value.Value;

      var normPlot = CreateScatterPlotResult(results, NormalizedPlotName);
      var plot = CreateScatterPlotResult(results, PlotName);
      foreach (var s in dataset.VariableNames) {
        if (!normPlot.Rows.ContainsKey(s)) normPlot.Rows.Add(new DataRow(s));
        if (!plot.Rows.ContainsKey(s)) plot.Rows.Add(new DataRow(s));

        if (dataset.Rows < lbsize) continue;
        var wd = dataset.GetDoubleValues(s, Enumerable.Range(dataset.Rows - lbsize, lbsize)).StandardDeviation();
        plot.Rows[s].Values.Add(wd);

        if (dataset.Rows < Math.Max(initialEvals, lbsize)) continue;
        var sd = dataset.GetDoubleValues(s, Enumerable.Range(0, initialEvals)).StandardDeviation();
        normPlot.Rows[s].Values.Add(wd / sd);
      }
      return base.Apply();
    }

    private static DataTable CreateScatterPlotResult(ResultCollection results, string plotname) {
      DataTable plot;
      if (!results.ContainsKey(plotname)) {
        plot = new DataTable(plotname) {
          VisualProperties = {
            XAxisTitle = "Iteration",
            YAxisTitle = plotname.Equals(NormalizedPlotName)? "Normalized Variance (Variance of last Samples)/(Variance of Initial Samples)" : "Standard deviation of last samples"
          }
        };
        results.Add(new Result(plotname, plot));
      }
      plot = (DataTable)results[plotname].Value;
      return plot;
    }

  }
}
